/* $Id: fish.c,v 1.20 1998/09/27 14:20:03 levitte Exp $
**++
**  FACILITY:	FISH main procedures
**
**  ABSTRACT:	SSH client
**
**  MODULE DESCRIPTION:
**
**  	This was a test client for NETLIB, that has been slammed together
**	together with an SSH client for Palm Pilot, resulting in FISH,
**	a SSH client for VMS.
**
**  AUTHOR OF TEST CLIENT:
**	 	    M. Madison
**  	    	    COPYRIGHT  1994, MADGOAT SOFTWARE.  ALL RIGHTS RESERVED.
**
**  AUTHORS OF FISH:
**		    Richard Levitte (RL) & Christer Weinigel (CW)
**
**  OTHER HACKERS:
**		    Wolfgang J. Moeller (WJM)
**
**  CREATION DATE:  
**
**  MODIFICATION HISTORY:
**
**  	1998-MAY-??	V0.0	CW	Initial coding.
**	1998-MAY-06	V0.1	RL	Quickened things up by moving some
**					things out of the AST routines.
** wjm 27-may-1998: (silently) fixed a few bugs relating to event flags.
** wjm 28-may-1998: netlib_read() for more than 1 byte does terminate
**		    at end of current packet (it *seems*).
**		    So raise read count (in AST only) to 64 = WJM_read_size
**		    (after CMU-IP TELNET).
**--
*/

#include "netlib_dir:netlibdef.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <descrip.h>
#include <climsgdef.h>
#include <clidef.h>
#include <stsdef.h>
#include <ssdef.h>
#include <iodef.h>
#include <lib$routines.h>
#include <ots$routines.h>
#include <str$routines.h>
#include <starlet.h>
#include <unistd.h>
#include <errno.h>
#ifdef __DECC
#include <builtins.h>
#else
#pragma builtins
#endif

#include "Pilot.h"

#include "ssh.h"
#include "ui.h"
#include "rand.h"
#include "erf.h"
#include "vms.h"
#include "util.h"

void hd(char *info, unsigned char *ptr, int count)
{
    char s[80], *p;
    int line, col;

    for (line = 0; line < count; line += 8)
    {
        p = s;
        p += sprintf(p, "%s", info);
        p += sprintf(p, "%04x  ", line);

        for (col = 0; col < 8; col++)
        {
            if (line + col < count)
                p += sprintf(p, "%02x ", ptr[line + col]);
            else
                p += sprintf(p, "   ");
        }

        *p++ = ' ';

        for (col = 0; col < 8; col++)
        {
            if (line + col >= count)
                break;

            if (ptr[line + col] >= ' ' && ptr[line + col] < 127)
                *p++ = ptr[line + col];
            else
                *p++ = '.';
        }

        *p = '\0';
        fprintf(stderr, "%s\n", s);
    }
}

#define INIT_SDESC(dsc, len, ptr) {(dsc).dsc$b_dtype = DSC$K_DTYPE_T;\
    (dsc).dsc$b_class = DSC$K_CLASS_S; (dsc).dsc$w_length = (len);\
    (dsc).dsc$a_pointer = (ptr);}

/*
**  Other statics we need because we have to pass these values by
**  reference
*/
static unsigned int sinsize = sizeof(struct SINDEF);
static unsigned int adrsize = sizeof(struct INADDRDEF);
static unsigned int which_dns = NETLIB_K_LOOKUP_DNS;
static unsigned int which_ht  = NETLIB_K_LOOKUP_HOST_TABLE;

/*
**  I/O request
*/
#define IORBUFSize 16384
typedef struct IORequest {
    struct IORequest *flink, *blink;
    struct NETLIBIOSBDEF iosb;
    struct dsc$descriptor bufdsc;
    char buf[IORBUFSize];
} IORequest;

#define WJM_read_size 64	/* taken from CMUIP TELNET */
#if WJM_read_size > IORBUFSize
	You lose ... !
#endif

/*
**  Queue
*/
typedef struct Queue {void *head, *tail;} Queue;

/*
**  Forward declarations
*/
unsigned int main(int argc, char *argv[]);
void Terminal_Read_AST(IORequest *ior);
void Network_Read_AST(IORequest *ior);
void Write_AST(IORequest *ior);
IORequest *Allocate_IOR(void);
IORequest *Get_IOR(void);

/*
**  OWN storage
*/
static Queue Free_Queue = {&Free_Queue, &Free_Queue};
char *input_device = "SYS$INPUT";
char *output_device = "SYS$OUTPUT";
static void *network_context;
static unsigned short network_ef;
static unsigned long network_ef_mask;
static unsigned char network_buffer[IORBUFSize];
static int network_buffer_n = 0;
static unsigned short input_ef;
static unsigned long input_ef_mask;
static unsigned char input_buffer[IORBUFSize];
static int input_buffer_n = -1;
unsigned short done_ef;
static unsigned long done_ef_mask;
static unsigned long ef_bits;
static unsigned int completion_status = SS$_NORMAL;
const unsigned int terminator_mask[2] = {0, 0};
static unsigned char timeout[8];
static unsigned char *timeout_ptr = 0;

/*
 * We're gonna change some of the characteristics of the terminal so I'll
 * declare a buffer for the terminal's characteristics before the changes
 * and another for after the changes. When the program exits the terminal's
 * characteristics will be reset the way they were before they were changed.
 */

tty_s *input_tty = 0, *output_tty = 0;


static void ssh_initprefs(ssh_prefs *p)
{
    char *cp;

    p->host = NULL;
    p->port = 22;
    p->username = xstrdup(getenv("USER"));
    for (cp = p->username; *cp; cp++)
	*cp = tolower(*cp);
    p->password = 0;
    p->ciphers = 0;		/* The user will specify later */
    p->ttymodes = NULL;
    p->ttymodelen = 0;
    p->commands = 0;
    p->infile = 0;
    p->inmode = 0;
    p->outfile = 0;
    p->outmode = 0;
    p->verbose = 0;
}

static void ssh_clearprefs(ssh_prefs *p)
{
    if (p->host) MemPtrFree(p->host);
    p->port = 0;
    if (p->username) MemPtrFree(p->username);
    if (p->ttymodes) MemPtrFree(p->ttymodes);
    ssh_initprefs(p);
}

void connected(ssh_state *state, Erf erf, void *erfp)
{
    IORequest *ior;
    unsigned int status;
    tty_s *ttyi = 0, *ttyo = 0;

    ssh_info("Entering interactive mode");
    if (state->sshprefs->verbose) {
	fprintf(stderr, "%%FISH-I-INFD, %d\n", state->infd);
	fprintf(stderr, "%%FISH-I-OUTFD, %d\n", state->outfd);
    }
    fprintf(stderr, "\n\n");

    input_buffer_n = 0;

    if (state->infd == 0) {
	ttyi = tty_open(input_device,&status);
	if (!tty_sensemode(ttyi,&status))
	    exit(status);
	input_tty = tty_dup(ttyi);
    }

    if (state->outfd == 0) {
	ttyo = tty_open(output_device,&status);
	if (!tty_sensemode(ttyo,&status))
	    exit(status);
	output_tty = tty_dup(ttyo);
    }

    if (ttyi) {
	tty_setflags(ttyi, /*TT$M_WRAP|*/ TT$M_NOECHO|TT$M_HOSTSYNC,TT2$M_PASTHRU,&status);
	tty_clearflags(ttyi,0 /*TT$M_TTSYNC*/ ,TT2$M_LOCALECHO,&status);
	if (!tty_setmode(ttyi,&status))
	    exit(status);
	tty_release(ttyi,&status);
    }

    if (ttyo) {
	tty_setflags(ttyo, /*TT$M_WRAP|*/ TT$M_NOECHO|TT$M_HOSTSYNC,TT2$M_PASTHRU,&status);
	tty_clearflags(ttyo,0 /*TT$M_TTSYNC*/ ,TT2$M_LOCALECHO,&status);
	if (!tty_setmode(ttyo,&status))
	    exit(status);
	tty_release(ttyo,&status);
    }

/*
**  Start a terminal read
*/
    if (state->infd == 0) {
	ior = Get_IOR();
	if (ior == 0) exit(SS$_INSFMEM);
	status = sys$qio(0, input_tty->channel,
			 IO$_READVBLK|IO$M_NOECHO|IO$M_NOFILTR,
			 &ior->iosb, Terminal_Read_AST, ior,
			 ior->buf, 1, 0, terminator_mask, 0, 0);
	if (!$VMS_STATUS_SUCCESS(status)) exit(status);
    }
}

static ssh_prefs gsshprefs;
static void *sshstate = NULL;

int sockfd;

static int gerrp;
Erf erf, erfp;

long sock_write(unsigned char *data, long len, void *state_,
		Erf erf, void *erfp)
{
    IORequest *ior;
    long totret = 0;
    unsigned int status;

    if (len < 0) {
	sys$setef(done_ef);

	totret = -1;
	goto sockwrite_out;
    }

    totret = len;

    while (len > 0) {
	int this_len = min(len, IORBUFSize);

/*
**  Write the data to the server.
*/
	ior = Get_IOR();
	if (ior == 0) status = SS$_INSFMEM;
	else
	{
	    ior->bufdsc.dsc$w_length = this_len;
	    memcpy(ior->buf, data, this_len);
	    status = netlib_write(&network_context, &ior->bufdsc,
				  0, 0, &ior->iosb, Write_AST, ior);
	}

	len -= this_len;

	if (!$VMS_STATUS_SUCCESS(status)) {
	    unsigned long ef_bits_to_check;
	    sys$readef(input_ef, &ef_bits_to_check);
	    if (!(ef_bits_to_check & done_ef_mask)) completion_status = status;
	    sys$setef(done_ef);
	    totret = -1;
	    goto sockwrite_out;
	}
    }

sockwrite_out:
    return totret;
}

/* This function is called when data arrives */
void ui_dispatch(unsigned char *data, long len, void *state_, Erf erf,
		 void *erfp)
{
    IORequest *ior;
    unsigned int status;

    if (len < 0) {
	sys$setef(done_ef);

	/* The connection was closed; pass on the info */
	goto ui_dispatch_out;
    }

    while (len > 0) {
	int this_len = min(len, IORBUFSize);

	ior = Get_IOR();
	if (ior == 0)
	    status = SS$_INSFMEM;
	else 
	{
	    memcpy(ior->buf, data, this_len);
	    
	    status = sys$qio(0, output_tty->channel, IO$_WRITEVBLK,
			     &ior->iosb, Write_AST, ior, ior->buf, this_len,
			     0, 0, 0, 0);
	}

	len -= this_len;

	if (!$VMS_STATUS_SUCCESS(status)) {
	    unsigned long ef_bits_to_check;
	    sys$readef(input_ef, &ef_bits_to_check);
	    if (!(ef_bits_to_check & done_ef_mask))
		completion_status = status;
	    sys$setef(done_ef);
	}
    }
	    
ui_dispatch_out:
    ;
}

void uf_dispatch(unsigned char *data, long len, void *state_, Erf erf,
		 void *erfp)
{
    int wr_status;

    if (len < 0) {
	sys$setef(done_ef);

	/* The connection was closed; pass on the info */
	goto uf_dispatch_out;
    }

    while(len > 0) {
	wr_status = write(((ssh_state *)sshstate)->outfd, data, len);
	if (wr_status < 0) {
	    close(((ssh_state *)sshstate)->outfd);
	    ((ssh_state *)sshstate)->outfd = -1;
	    sys$setef(done_ef);
	    goto uf_dispatch_out;
	}
	len -= wr_status;
	data += wr_status;
    }
	    
uf_dispatch_out:
    ;
}

/*
 *      e x i t _ h a n d l e r
 */
static void exit_handler ()
{
    int status;

    sys$setast (0);
    
    if (input_tty) {
	tty_io_cancel(input_tty, &status);
	tty_setmode(input_tty, &status);
	tty_close(input_tty, &status);
	input_tty = 0;
    }
    if (output_tty) {
	tty_io_cancel(output_tty, &status);
	tty_setmode(output_tty, &status);
	tty_close(output_tty, &status);
	output_tty = 0;
    }
}

void sighand(int signo)
{
    signal(SIGINT, SIG_DFL);

    fprintf(stderr, "Exiting on interrupt\n");

    exit(0);
}

/*
**  "Macro" to initialize a dynamic string descriptor.
*/
#define init_dyndesc(dsc) {\
	dsc.dsc$w_length = 0;\
	dsc.dsc$b_dtype = DSC$K_DTYPE_T;\
	dsc.dsc$b_class = DSC$K_CLASS_D;\
	dsc.dsc$a_pointer = 0;}

$DESCRIPTOR(cli_help,			"HELP");
$DESCRIPTOR(cli_version,		"VERSION");
$DESCRIPTOR(cli_host,			"HOST");
$DESCRIPTOR(cli_port,			"PORT");
$DESCRIPTOR(cli_password,		"PASSWORD");
$DESCRIPTOR(cli_username,		"USERNAME");
$DESCRIPTOR(cli_input,			"INPUT");
$DESCRIPTOR(cli_output,			"OUTPUT");
$DESCRIPTOR(cli_mode_input_text,	"MODE.INPUT.TEXT");
$DESCRIPTOR(cli_mode_input_binary,	"MODE.INPUT.BINARY");
$DESCRIPTOR(cli_mode_output_text,	"MODE.OUTPUT.TEXT");
$DESCRIPTOR(cli_mode_output_binary,	"MODE.OUTPUT.BINARY");
$DESCRIPTOR(cli_commands,		"COMMANDS");
$DESCRIPTOR(cli_ciphers,		"CIPHERS");
$DESCRIPTOR(cli_ciphers_none,		"CIPHERS.NONE");
$DESCRIPTOR(cli_ciphers_idea,		"CIPHERS.IDEA");
$DESCRIPTOR(cli_ciphers_3des,		"CIPHERS.3DES");
$DESCRIPTOR(cli_ciphers_des,		"CIPHERS.DES");
#if 0 /* Not supported */
$DESCRIPTOR(cli_ciphers_tss,		"CIPHERS.TSS");
#endif
$DESCRIPTOR(cli_ciphers_rc4,		"CIPHERS.RC4");
$DESCRIPTOR(cli_ciphers_blowfish,	"CIPHERS.BLOWFISH");
$DESCRIPTOR(cli_ciphers_all,		"CIPHERS.ALL");
$DESCRIPTOR(cli_auths,			"AUTHENTICATION_METHODS");
$DESCRIPTOR(cli_auths_all,		"AUTHENTICATION_METHODS.ALL");
$DESCRIPTOR(cli_auths_none,		"AUTHENTICATION_METHODS.NONE");
$DESCRIPTOR(cli_auths_rhosts,	       	"AUTHENTICATION_METHODS.RHOSTS");
$DESCRIPTOR(cli_auths_rsa,		"AUTHENTICATION_METHODS.RSA");
$DESCRIPTOR(cli_auths_password,		"AUTHENTICATION_METHODS.PASSWORD");
$DESCRIPTOR(cli_auths_rhosts_rsa,	"AUTHENTICATION_METHODS.RHRSA");
$DESCRIPTOR(cli_auths_tis,		"AUTHENTICATION_METHODS.TIS");
$DESCRIPTOR(cli_auths_kerberos,		"AUTHENTICATION_METHODS.KERBEROS");
$DESCRIPTOR(cli_verbose,		"VERBOSE");
$DESCRIPTOR(cli_verbose_full,		"VERBOSE.FULL");
$DESCRIPTOR(cli_verbose_password,	"VERBOSE.PASSWORD");

$DESCRIPTOR(fish_command,		"FISH ");

#ifdef __DECC
#pragma extern_model save
#pragma extern_model strict_refdef
extern void *fish_clitables;
#pragma extern_model restore
#else
globalref void *fish_clitables;
#endif

extern unsigned long cli$dcl_parse ();
extern unsigned long cli$present ();
extern unsigned long cli$get_value ();

static void usage()
{
    fprintf(stderr, "FISH [qualifier ...] host [port]\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Available qualifiers:\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t/HELP         gives this message\n");
    fprintf(stderr, "\t/VERSION      gives version number\n");
    fprintf(stderr, "\t/VERBOSE      writes some extra info\n");
    fprintf(stderr, "\t/COMMANDS=s   commands to execute instead of going interactive\n");
    fprintf(stderr, "\t/INPUT=file   input redirection\n");
    fprintf(stderr, "\t/OUTPUT=file  output redirection\n");
    fprintf(stderr, "\t/MODE=modes   redirection file modes\n");
    fprintf(stderr, "\t/USERNAME=u   the username you want to use on the remote host\n");
    fprintf(stderr, "\t/PASSWORD=p   the password you want to use on the remote host.  DANGEROUS!\n");
    fprintf(stderr, "\t/CIPHERS=c    One or several ciphers you want to use:\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t              All, IDEA, 3DES, DES, RC4, Blowfish\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\tAUTHENTICATION_METHODS\n");
    fprintf(stderr, "\t              One or several authentication methods to try:\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t              All, RHOSTS, RSA, RHRSA, PASSWORD, TIS, Kerberos\n");
    fprintf(stderr, "\t              (only RHOSTS, PASSWORD and TIS are currently\n");
    fprintf(stderr, "\t              implemented)\n");
    fprintf(stderr, "\n");
}

static unsigned long
check_cli (struct dsc$descriptor_s *qual)
{
/*
**  Routine:	check_cli
**
**  Function:	Check to see if a CLD was used to invoke the program.
**
*/
    lib$establish(lib$sig_to_ret);	/* Establish condition handler */
    return (cli$present(qual));		/* Just see if something was given */
}

static void unquote(char *s)
{
    char *p1,*p2;
    for (p1=p2=s; *p2 != '\0'; ) {
	if (*p2 == '"') {
	    p2++;
	    if (*p2 == '"') {
		*p1++ = '"';
		p2++;
	    }
	} else {
	    *p1++ = *p2++;
	}
    }
    *p1 = *p2;
}

unsigned int parse_args(ssh_prefs *sshprefs)
{
    int status, h_status = 0, v_status = 0;
    struct dsc$descriptor_d work_str;

    init_dyndesc(work_str);

    /*
    **  See if the program was invoked by the CLI (SET COMMAND) or by
    **  a foreign command definition.  Check for HOST, which is a
    **  required parameter in any case.
    */
    status = check_cli (&cli_host);
    if (!(status & 1)) {
	lib$get_foreign (&work_str);
	str$concat (&work_str, &fish_command, &work_str);
	status = cli$dcl_parse(&work_str, &fish_clitables, lib$get_input,
			lib$get_input, 0);
	if (!(status & 1))
	    return (unsigned int)-1; /* Because the error has already
					been reported */
    }

    status = cli$present(&cli_verbose);
    if (status == CLI$_PRESENT) {
	gsshprefs.verbose = 1;
    }
    if (status == CLI$_NEGATED) {
	gsshprefs.verbose = 0;
    }

    if (status & 1)
    {
	status = cli$present(&cli_verbose_full);
	if (status & 1)
	    gsshprefs.verbose |= 2;
	status = cli$present(&cli_verbose_password);
	if (status & 1)
	    gsshprefs.verbose |= 4;
    }

    h_status = cli$present(&cli_help);
    v_status = cli$present(&cli_version);
    if (h_status & 1)
	if (v_status & 1)
	    fprintf(stderr, "%%FISH-I-IGNORE, /HELP and /VERSION selected, all other options will be ignored\n");
	else
	    fprintf(stderr, "%%FISH-I-IGNORE, /HELP selected, all other options will be ignored\n");
    else if (v_status & 1)
	fprintf(stderr, "%%FISH-I-IGNORE, /VERSION selected, all other options will be ignored\n");

    if (h_status & 1)
	usage();

    if (v_status & 1)
	fprintf(stderr, "%%FISH-I-VERSION, FISH version %s\n", FISH_VERSION);

    if ((h_status & 1) || (v_status & 1))
	return (unsigned int)-1;

    status = cli$present(&cli_host);
    if (status & 1) {
	status = cli$get_value(&cli_host, &work_str);
	gsshprefs.host = xstrndup(work_str.dsc$a_pointer,
				  work_str.dsc$w_length);
	unquote(gsshprefs.host);
    }

    status = cli$present(&cli_port);
    if (status & 1) {
	status = cli$get_value(&cli_port, &work_str);
	status = ots$cvt_tu_l (&work_str, &gsshprefs.port,
			       sizeof (gsshprefs.port),0x11);
    }

    status = cli$present(&cli_username);
    if (status & 1) {
	register unsigned char *cp;

	status = cli$get_value(&cli_username, &work_str);
	gsshprefs.username = xstrndup(work_str.dsc$a_pointer,
				  work_str.dsc$w_length);
	if (*gsshprefs.username == '"')
	    unquote(gsshprefs.username);
	else
	    for (cp = gsshprefs.username; *cp; cp++)
		*cp = tolower(*cp);
    }

    status = cli$present(&cli_password);
    if (status & 1) {
	register unsigned char *cp;

	status = cli$get_value(&cli_password, &work_str);
	gsshprefs.password = xstrndup(work_str.dsc$a_pointer,
				  work_str.dsc$w_length);
	unquote(gsshprefs.password);
    }

    status = cli$present(&cli_ciphers);
    if (status & 1) {
	int all_ciphers = cli$present(&cli_ciphers_all) & 1;
	int wanted_ciphers = 1 << SSH_CIPHER_NONE;

	status = cli$present(&cli_ciphers_des);
	if (status == CLI$_PRESENT || all_ciphers)
	    wanted_ciphers |= 1 << SSH_CIPHER_DES;
	if (status == CLI$_NEGATED)
	    wanted_ciphers &= ~(1 << SSH_CIPHER_DES);

	status = cli$present(&cli_ciphers_3des);
	if (status == CLI$_PRESENT || all_ciphers)
	    wanted_ciphers |= 1 << SSH_CIPHER_3DES;
	if (status == CLI$_NEGATED)
	    wanted_ciphers &= ~(1 << SSH_CIPHER_3DES);

	status = cli$present(&cli_ciphers_idea);
	if (status == CLI$_PRESENT || all_ciphers)
	    wanted_ciphers |= 1 << SSH_CIPHER_IDEA;
	if (status == CLI$_NEGATED)
	    wanted_ciphers &= ~(1 << SSH_CIPHER_IDEA);

	status = cli$present(&cli_ciphers_rc4);
	if (status == CLI$_PRESENT || all_ciphers)
	    wanted_ciphers |= 1 << SSH_CIPHER_RC4;
	if (status == CLI$_NEGATED)
	    wanted_ciphers &= ~(1 << SSH_CIPHER_RC4);

#if 0 /* Not currently supported */
	status = cli$present(&cli_ciphers_tss);
	if (status == CLI$_PRESENT || all_ciphers)
	    wanted_ciphers |= 1 << SSH_CIPHER_TSS;
	if (status == CLI$_NEGATED)
	    wanted_ciphers &= ~(1 << SSH_CIPHER_TSS);
#endif

	status = cli$present(&cli_ciphers_blowfish);
	if (status == CLI$_PRESENT || all_ciphers)
	    wanted_ciphers |= 1 << SSH_CIPHER_BLOWFISH;
	if (status == CLI$_NEGATED)
	    wanted_ciphers &= ~(1 << SSH_CIPHER_BLOWFISH);

	gsshprefs.ciphers = wanted_ciphers;
    } else {
	gsshprefs.ciphers = 1 << SSH_CIPHER_NONE;
	gsshprefs.ciphers |= 1 << SSH_CIPHER_DES;
	gsshprefs.ciphers |= 1 << SSH_CIPHER_3DES;
	gsshprefs.ciphers |= 1 << SSH_CIPHER_IDEA;
	gsshprefs.ciphers |= 1 << SSH_CIPHER_RC4;
	gsshprefs.ciphers |= 1 << SSH_CIPHER_TSS;
	gsshprefs.ciphers |= 1 << SSH_CIPHER_BLOWFISH;
    }

    status = cli$present(&cli_auths);
    if (status & 1) {
	int all_auths = cli$present(&cli_auths_all) & 1;
	int wanted_auths = 1 << SSH_AUTH_NONE;

	status = cli$present(&cli_auths_rhosts);
	if (status == CLI$_PRESENT || all_auths)
	    wanted_auths |= 1 << SSH_AUTH_RHOSTS;
	if (status == CLI$_NEGATED)
	    wanted_auths &= ~(1 << SSH_AUTH_RHOSTS);

#if 0
	status = cli$present(&cli_auths_rsa);
	if (status == CLI$_PRESENT || all_auths)
	    wanted_auths |= 1 << SSH_AUTH_RSA;
	if (status == CLI$_NEGATED)
	    wanted_auths &= ~(1 << SSH_AUTH_RSA);

	status = cli$present(&cli_auths_rhosts_rsa);
	if (status == CLI$_PRESENT || all_auths)
	    wanted_auths |= 1 << SSH_AUTH_RHOSTS_RSA;
	if (status == CLI$_NEGATED)
	    wanted_auths &= ~(1 << SSH_AUTH_RHOSTS_RSA);
#endif

	status = cli$present(&cli_auths_password);
	if (status == CLI$_PRESENT || all_auths)
	    wanted_auths |= 1 << SSH_AUTH_PASSWORD;
	if (status == CLI$_NEGATED)
	    wanted_auths &= ~(1 << SSH_AUTH_PASSWORD);

	status = cli$present(&cli_auths_tis);
	if (status == CLI$_PRESENT || all_auths)
	    wanted_auths |= 1 << SSH_AUTH_TIS;
	if (status == CLI$_NEGATED)
	    wanted_auths &= ~(1 << SSH_AUTH_TIS);

#if 0
	status = cli$present(&cli_auths_kerberos);
	if (status == CLI$_PRESENT || all_auths)
	    wanted_auths |= 1 << SSH_AUTH_KERBEROS;
	if (status == CLI$_NEGATED)
	    wanted_auths &= ~(1 << SSH_AUTH_KERBEROS);
#endif

	gsshprefs.auths = wanted_auths;
    } else {
	gsshprefs.auths = 1 << SSH_AUTH_NONE;
	gsshprefs.auths |= 1 << SSH_AUTH_RHOSTS;
	gsshprefs.auths |= 1 << SSH_AUTH_RSA;
	gsshprefs.auths |= 1 << SSH_AUTH_RHOSTS_RSA;
	gsshprefs.auths |= 1 << SSH_AUTH_PASSWORD;
	gsshprefs.auths |= 1 << SSH_AUTH_TIS;
	gsshprefs.auths |= 1 << SSH_AUTH_KERBEROS;
    }

    gsshprefs.auths &=
	(1 << SSH_AUTH_NONE) | (1 << SSH_AUTH_RHOSTS)
	    | (1 << SSH_AUTH_PASSWORD) | (1 << SSH_AUTH_TIS);

    status = cli$present(&cli_commands);
    if (status & 1) {
	status = cli$get_value(&cli_commands, &work_str);
	gsshprefs.commands = xstrndup(work_str.dsc$a_pointer,
				      work_str.dsc$w_length);
	unquote(gsshprefs.commands);
    }

    status = cli$present(&cli_input);
    if (status & 1) {
	char *p1, *p2;
	status = cli$get_value(&cli_input, &work_str);
	gsshprefs.infile = xstrndup(work_str.dsc$a_pointer,
				    work_str.dsc$w_length);
    }

    status = cli$present(&cli_output);
    if (status & 1) {
	char *p1, *p2;
	status = cli$get_value(&cli_output, &work_str);
	gsshprefs.outfile = xstrndup(work_str.dsc$a_pointer,
				     work_str.dsc$w_length);
    }

    status = cli$present(&cli_mode_input_text);
    if (status & 1)
	gsshprefs.inmode = 0;
    status = cli$present(&cli_mode_input_binary);
    if (status & 1)
	gsshprefs.inmode = 1;
    status = cli$present(&cli_mode_output_text);
    if (status & 1)
	gsshprefs.outmode = 0;
    status = cli$present(&cli_mode_output_binary);
    if (status & 1)
	gsshprefs.outmode = 1;

    return 1;
}

unsigned int main(int argc, char *argv[])
{
    char randfilename[32];
    IORequest *ior;
    struct SINDEF remsin;
    struct dsc$descriptor dsc;
#ifndef WJM_read_size
    struct dsc$descriptor dsc_timeout;
    char net_time[] = "0 00:00:00.20";
#endif
    unsigned int status, remlen;
    unsigned short port, namelen;
    char name[1024], buf[1024];
    int i;
    static int excode;
    struct NETLIBIOSBDEF iosb;
    struct exit_handler_block 
    {
        struct exit_handler_block *flink;
        void (*exit_routine)();
        int arg_count;
        int *status_address;
        int exit_status;
    } exhblk = {0, exit_handler, 1, &excode, 0};

    ssh_initprefs(&gsshprefs);

    status = parse_args(&gsshprefs);
    if (status == (unsigned int)-1)
	return SS$_NORMAL;
    if (!(status & 1))
	return status;

    /* Load up with some randomness; use the saved file and the memopad */

    RAND_file_name(randfilename, 32);
    RAND_load_file(randfilename, 0x7fffffffL);

/*
** Set up exit handler
*/

    status = sys$dclexh (&exhblk);
    if (!$VMS_STATUS_SUCCESS(status)) return(status);

    if (gsshprefs.infile != 0)
	signal(SIGINT, sighand);

/*
**  Allocate a network socket and connect to the server
*/
    status = netlib_socket(&network_context);
    if (!$VMS_STATUS_SUCCESS(status)) return status;

    INIT_SDESC(dsc, strlen(gsshprefs.host), gsshprefs.host);
    port = gsshprefs.port;
    status = netlib_connect_by_name(&network_context, &dsc, &port, &iosb);
    if (status == 0x870
	|| iosb.iosb_w_status == 0x870) /* SYSTEM-W-ENDOFFILE */
    {
	fprintf(stderr, "%%FISH-E-HOSTNOTFOUND, Host not found, or slow DNS lookup\n");
	return SS$_NORMAL;	/* XXX: Oh, I guess it'll soon be time to build
				   a message file...  After the weekend :-) */
    }
    if (!$VMS_STATUS_SUCCESS(iosb.iosb_w_status))
	return iosb.iosb_w_status;
    if (!$VMS_STATUS_SUCCESS(status)) return status;

/*
**  Allocate event flags for signaling to the main loop.
*/
    status = lib$get_ef(&input_ef);
    if (!$VMS_STATUS_SUCCESS(status)) return status;
    input_ef_mask = 1<<(input_ef & 31);
    status = sys$clref(input_ef);
    if (!$VMS_STATUS_SUCCESS(status)) return status;

    status = lib$get_ef(&network_ef);
    if (!$VMS_STATUS_SUCCESS(status)) return status;
    network_ef_mask = 1<<(network_ef & 31);
    status = sys$clref(network_ef);
    if (!$VMS_STATUS_SUCCESS(status)) return status;

    status = lib$get_ef(&done_ef);
    if (!$VMS_STATUS_SUCCESS(status)) return status;
    done_ef_mask = 1<<(done_ef & 31);
    status = sys$clref(done_ef);
    if (!$VMS_STATUS_SUCCESS(status)) return status;

    ef_bits = input_ef_mask | network_ef_mask | done_ef_mask;

#ifndef WJM_read_size
/* Setup timeout value for network reads */
 
    INIT_SDESC(dsc_timeout, strlen(net_time), net_time);
    status = sys$bintim(&dsc_timeout, &timeout);
    if (!$VMS_STATUS_SUCCESS(status)) return status;
#endif

/*
**  Let the user know who we're connected to
*/
    status = netlib_getpeername(&network_context, &remsin, &sinsize, &remlen);
    if ($VMS_STATUS_SUCCESS(status)) {
    	INIT_SDESC(dsc, sizeof(name), name);
    	status = netlib_address_to_name(&network_context, &which_dns,
					&remsin.sin_x_addr, &adrsize,
					&dsc, &namelen);
    	if (!$VMS_STATUS_SUCCESS(status))
    	    status = netlib_address_to_name(&network_context, &which_ht,
					    &remsin.sin_x_addr, &adrsize,
					    &dsc, &namelen);
    	if (!$VMS_STATUS_SUCCESS(status))
    	    status = netlib_addrtostr(&remsin.sin_x_addr, &dsc, &namelen);
    	if ($VMS_STATUS_SUCCESS(status)) {
    	    i = sprintf(buf, "Connected to host %.*s, port %d.",
			namelen, name, netlib_ntoh_word(&remsin.sin_w_port));
    	    INIT_SDESC(dsc, i, buf);
    	    lib$put_output(&dsc);
    	    dsc.dsc$w_length = 0;
    	    lib$put_output(&dsc);
    	}
    }

/*
**  Pre-load the free queue
*/
    for (i = 0; i < 16; i++) {
    	ior = Allocate_IOR();
    	if (ior == 0) break;
    	_INSQUE(ior, Free_Queue.tail);
    }

    /* Set up an SSH connection */
    sshstate = ssh_init(sock_write, NULL, erf, erfp);
    if (!sshstate) exit(1);

    /* Set the preferences */
    ssh_set_prefs(&gsshprefs, sshstate, erf, erfp);

    {
	char *termname = getenv("TERM");

	if (termname)
	{
	    char *t2 = xstrdup(termname);
	    char *p;

	    if ((p = strchr(t2, '-')) != NULL)
		*p = '\0';

	    input_tty = tty_open(input_device, &status);
	    if (!tty_sensemode(input_tty,&status))
		exit(status);
	    ssh_setterm(t2, (input_tty->tty_char.basic_chars >> 24),
			input_tty->tty_char.buffer_size,
			0, 0, sshstate, erf, erfp);
	    tty_close(input_tty,&status);
	    input_tty = 0;
	    
	    free(t2);
	}
    }

    /* Set the dispatch path */
    if (((ssh_state *)sshstate)->outfd == 0)
	ssh_set_dispatch(ui_dispatch, NULL, sshstate, erf, erfp);
    else
	ssh_set_dispatch(uf_dispatch, NULL, sshstate, erf, erfp);

#ifndef WJM_read_size
    ior->bufdsc.dsc$w_length = 1; /* Initial read is for one byte, untimed */
#endif

/*
**  Start a network read
*/
    ior = Get_IOR();
    if (ior == 0) return SS$_INSFMEM;
    status = netlib_read(&network_context, &ior->bufdsc,
    	    	    	  0, 0, 0, 0, &ior->iosb,
    	    	    	  Network_Read_AST, ior);
    if (!$VMS_STATUS_SUCCESS(status)) {
    	sys$cancel(input_tty->channel);
    	return SS$_INSFMEM;
    }

/*
**  Let the ASTs do most of the work.  They will set appropriate event
**  flags when it's time for this loop to do some external work, or when
**  it's time to shut down.
*/
    while (1) {
	unsigned long ef_bits_to_check;
	if (((ssh_state *)sshstate)->infd <= 0)
	    status = sys$wflor(input_ef,ef_bits);
	status = sys$readef(input_ef,&ef_bits_to_check);
	if (((ssh_state *)sshstate)->infd == 0) {
	    if (ef_bits_to_check & input_ef_mask) {
		sys$setast(0);
/*
**  Send data downlink through SSH layer
*/
		ssh_write(input_buffer, input_buffer_n, sshstate, erf, &gerrp);
		input_buffer_n = 0;
		sys$clref(input_ef);
		sys$setast(1);
	    }
	} else if (((ssh_state *)sshstate)->infd > 0 && input_buffer_n >= 0) {
	    input_buffer_n = read(((ssh_state *)sshstate)->infd,
				  input_buffer, IORBUFSize);
	    if (((ssh_state *)sshstate)->sshprefs->verbose)
		fprintf(stderr, "%%FISH-I-FILEINPUT, [%d/%d] \"%-*.*s\"\n",
			input_buffer_n, IORBUFSize,
			input_buffer_n, input_buffer_n, input_buffer);
	    if (input_buffer_n == 0) {
		close(((ssh_state *)sshstate)->infd);
		((ssh_state *)sshstate)->infd = -1;
		ssh_write_type(SSH_CMSG_EOF, NULL, 0, (ssh_state *)sshstate,
			       erf, erfp);
	    } else if (input_buffer_n < 0) {
#ifndef EAGAIN
#define EAGAIN -1
#endif
#ifndef EWOULDBLOCK
#define EWOULDBLOCK -1
#endif
		if (errno != EAGAIN && errno != EWOULDBLOCK) {
		    close(((ssh_state *)sshstate)->infd);
		    ((ssh_state *)sshstate)->infd = -1;
		}
	    } else
		ssh_write(input_buffer, input_buffer_n, sshstate, erf, &gerrp);
	}
	if (ef_bits_to_check & done_ef_mask) {
	    break;
	}
    }

/*
**  Close down the network socket and terminal channels.
*/
    netlib_shutdown(&network_context);
    netlib_close(&network_context);
    exit_handler();

    /* Save our randomness state */
    RAND_file_name(randfilename, 32);
    RAND_write_file(randfilename);

    ssh_clearprefs(&gsshprefs);

    return completion_status;
}



/*
**++
**  ROUTINE:	Terminal_Read_AST
**
**  FUNCTIONAL DESCRIPTION:
**
**  	Read completion AST for terminal input.
**
**  RETURNS:	void
**
**  PROTOTYPE:
**
**  	Terminal_Read_AST(ior)
**
**  IMPLICIT INPUTS:	None.
**
**  IMPLICIT OUTPUTS:	None.
**
**  COMPLETION CODES:	None.
**
**  SIDE EFFECTS:   	None.
**
**--
*/
void Terminal_Read_AST (IORequest *ior) {

    unsigned int status;
    static enum { NOTHING, CR_SEEN, TILDE_SEEN } keystate = NOTHING;
    register unsigned int c = -1;
    unsigned long ef_bits_to_check;

    if (ior->iosb.iosb_w_count != 0) {
/*
**  Exit on error
*/
	if (!$VMS_STATUS_SUCCESS(ior->iosb.iosb_w_status)) {
	    sys$readef(input_ef, &ef_bits_to_check);
	    if (!(ef_bits_to_check & done_ef_mask))
		completion_status = ior->iosb.iosb_w_status;
	    sys$setef(done_ef);
	    return;
	}

	c = ior->buf[0];
    }

/*
**  Start a new read
*/
    status = SS$_NORMAL;
    sys$readef(input_ef, &ef_bits_to_check);
    if (!(ef_bits_to_check & done_ef_mask))
        status = sys$qio(0, input_tty->channel,
			 IO$_READVBLK|IO$M_NOECHO|IO$M_NOFILTR,
			 &ior->iosb, Terminal_Read_AST, ior,
			 ior->buf, 1, 0, terminator_mask, 0, 0);
    if (!$VMS_STATUS_SUCCESS(status)) {
	unsigned long ef_bits_to_check;
	sys$readef(input_ef, &ef_bits_to_check);
	if (!(ef_bits_to_check & done_ef_mask))
	    completion_status = ior->iosb.iosb_w_status;
	sys$setef(done_ef);
    	return;
    }

    if (c >= 0) {
	switch (keystate)
	{
	case NOTHING:
	    if (c == '\r' || c == '\n')
		keystate = CR_SEEN;
	    break;
	case CR_SEEN:
	    if (c == '~')
		keystate = TILDE_SEEN;
	    else
		keystate = NOTHING;
	    break;
	case TILDE_SEEN:
	    if (c == '.')
	    {
		ssh_fatal("Connection closed");
	    }
	    else
		keystate = NOTHING;
	    break;
	}
	input_buffer[input_buffer_n++] = c;
	sys$setef(input_ef);
    }
} /* Terminal_Read_AST */

/*
**++
**  ROUTINE:	Network_Read_AST
**
**  FUNCTIONAL DESCRIPTION:
**
**  	Completion AST for network read.
**
**  RETURNS:	void
**
**  PROTOTYPE:
**
**  	Network_Read_AST(ior)
**
**  IMPLICIT INPUTS:	None.
**
**  IMPLICIT OUTPUTS:	None.
**
**  COMPLETION CODES:	None.
**
**  SIDE EFFECTS:   	None.
**
**--
*/
void Network_Read_AST (IORequest *ior) {

    unsigned int status;
    int len;

/*
**  Exit on error
*/
    if (!$VMS_STATUS_SUCCESS(ior->iosb.iosb_w_status)
#ifndef WJM_read_size
	&& ior->iosb.iosb_w_status != SS$_TIMEOUT
#endif
	) {
	unsigned long ef_bits_to_check;
	sys$readef(input_ef, &ef_bits_to_check);
    	if (!(ef_bits_to_check & done_ef_mask))
	    completion_status = ior->iosb.iosb_w_status;
	sys$setef(done_ef);
    	return;
    }

/* 
**  Send data up through ssh layer 
*/  
#ifdef WJM_read_size
    ssh_dispatch(ior->buf, ior->iosb.iosb_w_count, sshstate, erf, erfp);
#else	/* WJM_read_size */
    len = ior->iosb.iosb_w_count;
    if (len > 0)
	ssh_dispatch(ior->buf, len, sshstate, erf, erfp);
#endif	/* WJM_read_size */
    
/*
**  Start a new read from the network
*/
#ifdef WJM_read_size
    ior->bufdsc.dsc$w_length = WJM_read_size;
#else /* WJM_read_size */
/*
** If the previous read timed out, we start an untimed read for 1 byte.
** Otherwise, start a timed read for the full buffer.
*/
    if (status == SS$_TIMEOUT) {
	ior->bufdsc.dsc$w_length = 1;
	timeout_ptr = 0;
    } else {
	ior->bufdsc.dsc$w_length = IORBUFSize;
	timeout_ptr = &timeout;
    }
#endif	/* WJM_read_size */
    status = netlib_read(&network_context, &ior->bufdsc,
			 0, 0, 0, timeout_ptr, &ior->iosb,
			 Network_Read_AST, ior);
    if (!$VMS_STATUS_SUCCESS(status)) {
	unsigned long ef_bits_to_check;
	sys$readef(input_ef, &ef_bits_to_check);
    	if (!(ef_bits_to_check & done_ef_mask))
	    completion_status = ior->iosb.iosb_w_status;
	sys$setef(done_ef);
    	return;
    }
} /* Network_Read_AST */

unsigned long ior_allocation_count = 0;	/* Statistics */
unsigned long ior_use_count = 0;	/* Statistics */

/*
**++
**  ROUTINE:	Write_AST
**
**  FUNCTIONAL DESCRIPTION:
**
**  	Completion AST for terminal and network writes.  It checks
**  for successful completion and frees the I/O request block.
**
**  RETURNS:	void
**
**  PROTOTYPE:
**
**  	Write_AST(ior)
**
**  IMPLICIT INPUTS:	None.
**
**  IMPLICIT OUTPUTS:	None.
**
**  COMPLETION CODES:	None.
**
**  SIDE EFFECTS:   	None.
**
**--
*/
void Write_AST (IORequest *ior) {

    if (!$VMS_STATUS_SUCCESS(ior->iosb.iosb_w_status)) {
	unsigned long ef_bits_to_check;
	sys$readef(input_ef, &ef_bits_to_check);
    	if (!(ef_bits_to_check & done_ef_mask))
	    completion_status = ior->iosb.iosb_w_status;
	sys$setef(done_ef);
    	return;
    }

    _INSQUE(ior, Free_Queue.tail);
    ior_use_count--;


} /* Write_AST */

/*
**++
**  ROUTINE:	Allocate_IOR
**
**  FUNCTIONAL DESCRIPTION:
**
**  	Allocates an IORequest.
**
**  RETURNS:    IORequest *
**
**  PROTOTYPE:
**
**  	Allocate_IOR()
**
**  IMPLICIT INPUTS:	None.
**
**  IMPLICIT OUTPUTS:	None.
**
**  COMPLETION CODES:	0 = allocation failure
**
**  SIDE EFFECTS:   	None.
**
**--
*/
IORequest *Allocate_IOR (void) {

    IORequest *ior;
    unsigned int status;
    const unsigned int iorsize = sizeof(IORequest);

    status = lib$get_vm(&iorsize, &ior);
    if (!$VMS_STATUS_SUCCESS(status)) return 0;

    memset(ior, 0, iorsize);
    INIT_SDESC(ior->bufdsc, 1, ior->buf);

    ior_allocation_count++;

    return ior;

} /* Allocate_IOR */

/*
**++
**  ROUTINE:	Get_IOR
**
**  FUNCTIONAL DESCRIPTION:
**
**  	Gets a free IORequest, either off the free queue or by
**  allocating one.
**
**  RETURNS:	IORequest *
**
**  PROTOTYPE:
**
**  	Get_IOR()
**
**  IMPLICIT INPUTS:	None.
**
**  IMPLICIT OUTPUTS:	None.
**
**  COMPLETION CODES:	0 = couldn't get one
**
**  SIDE EFFECTS:   	None.
**
**--
*/
IORequest *Get_IOR (void) {

    IORequest *ior;

    /* Let's watch out, it looks like things get a little overwritten if
       we don't do the extra check on Free_Queue.head... */
    if (Free_Queue.head == &Free_Queue
	/* Only return code 2 means there was nothing in the queue... */
	|| _REMQUE(Free_Queue.head, (void **) &ior) == 2) 
	ior = Allocate_IOR();

    if (ior)
	ior->bufdsc.dsc$w_length = WJM_read_size;

    ior_use_count++;

    return ior;
} /* Get_IOR */

/* Emacs local variables

Local variables:
eval: (set-c-style "BSD")
end:

*/
